<!doctype html>
<head>
  <meta charset="utf-8">
  <meta name="description" content="{{ site.description }}">
  <meta name="author" content="{{ site.author }}">
  <meta name="viewport" content="width=device-width,initial-scale=1">
  <title>{{ site.title }}{% if page.title %} &mdash; {{ page.title }}{% endif %}</title>
  {% if site.css %}
  <link rel="stylesheet" type="text/css" href="{{ site.root }}{{ site.css }}">
  {% else %}
  <style type="text/css">
body {
    font-family: 'Baskerville', Georgia, Times, serif;
    font-size: 1.3em;
    line-height: 1.5em;
    padding: 0 1.5em;
    margin: 0;
}
h1, h2, h3, h4, h5, h6 { font-weight: bold; }
h1 { font-size: 2em; line-height: 0.75em; margin: 0.75em 0; }
h2 { font-size: 1.5em; line-height: 1em; margin: 1em 0; }
h3 { font-size: 1.2em; line-height: 1.25em; margin: 1.25em 0; }
h4 { font-size: 1em; line-height: 1.5em; margin: 1.5em 0; }
h5 { font-size: 0.8em; line-height: 1.875em; margin: 1.875em 0; }
h6 { font-size: 0.75em; line-height: 2em; margin: 2em 0; }
p {
    font-size: 1em;
    line-height: 1.5em;
    margin: 1.5em 0;
}
code { background: #fee; padding: 0 0.3em; }
#container > header {
    text-align: center;
}
#container > footer {
    text-align: center;
    font-style: italic;
    font-size: 0.8em;
    color: #888;
}

.site-title {
    font-size: 2em;
    font-weight: bold;
    line-height: 0.75em; 
    margin: 0.75em 0; 
}
.post h2 { margin-bottom: 0.33em; }
.date {
    font-style: italic;
    font-size: 0.75em;
    line-height: 0.5em;
    margin-bottom: 1.5em;
}
blockquote {
    font-style: italic;
}
  </style>
  {% endif %}
</head>
<body>
  <div id="container">
    <header>
        {% block header %}
        {% if page.title %}
        <p class="site-title"><a href="{{ site.root }}">{{ site.title }}</a></p>
        {% else %}
        <h1><a href="{{ site.root }}">{{ site.title }}</a></h1>
        {% endif %}
        {% endblock %}
    </header>
    <section id="main" role="main">
        {% block main %}
        {% if page.title %}
        <h1 class="page-title">{{ page.title }}</h1>
        {% endif %}
        {% if page.tags %}
        <div class="tags">
            <span>Posted in</span>
        {% for t in page.tags -%}
            <span class="tag"><a href="{{ pctagurl(t) }}">{{t}}</a>{% if not loop.last %}, {% endif %}</span>
        {%- endfor %}
        </div>
        {% endif %}

        {{ content|safe }}
        {% endblock %}
    </section>
    <footer>
        {% block footer %}
        {% endblock %}
        {{ piecrust.debug_info|safe }}
        <p>{{ piecrust.branding|safe }}</p>
    </footer>
  </div>
</body>
</html>
